print("import: 'robot'")
import robot

print("import: 'robot.api'")
import robot.api

print("import: 'robot.conf'")
import robot.conf

print("import: 'robot.htmldata'")
import robot.htmldata

print("import: 'robot.libdocpkg'")
import robot.libdocpkg

print("import: 'robot.libraries'")
import robot.libraries

print("import: 'robot.model'")
import robot.model

print("import: 'robot.output'")
import robot.output

print("import: 'robot.output.console'")
import robot.output.console

print("import: 'robot.parsing'")
import robot.parsing

print("import: 'robot.reporting'")
import robot.reporting

print("import: 'robot.result'")
import robot.result

print("import: 'robot.running'")
import robot.running

print("import: 'robot.running.arguments'")
import robot.running.arguments

print("import: 'robot.running.timeouts'")
import robot.running.timeouts

print("import: 'robot.utils'")
import robot.utils

print("import: 'robot.variables'")
import robot.variables

